/*
 *   LdSerLib based on File-X-Fer - based on SimpleTerm
 *
 *   This program is used for loading the serial library. Some old
 *   palm applications rely on this library beeing preloaded at boot time.
 *
 *   Copyright (C) 2002 by gru <gru0@onlinehome.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define NON_PORTABLE

#include <Pilot.h>
#include <SerialMgr.h>
#include <SystemPrv.h>
#include <SysEvtMgr.h>

#include "ldserlib.h"
#include "resource.h"

/*
 * structres
 */
typedef struct {
        UInt version;
	UInt baud;
	UInt bits;
	UInt parity;
	UInt xon;
	UInt rts;
} LdSerLibPreferenceType;

/*
 * defines
 */
#define versionRequired 0x02000000

#define idAppLdSerLib 'gru0'
#define noFormLoaded -1

/*
 * prototypes for internal functions
 */
// application stuff
static Err get_or_create_prefs(void);
static Err StartApplication(void);
static void StopApplication(void);
static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags);

// database routines
// open
static void GetSizeString(CharPtr label, ULong Size, Boolean longFormat);

// form routines
static Boolean FormHandleMenuEvent(EventPtr event);
// main form
static void MainFormInit(FormPtr frmLdSerLib);
static void MainFormTableScroll(FormPtr frmLdSerLib, Int linesToScroll, Boolean relative);
static void MainFormDrawTable(VoidPtr tablePtr, Word row, Word column, RectanglePtr bounds);
static Boolean MainFormHandleEvent(EventPtr event);
// preferences form
static void InitPreferences(FormPtr frm);
static Boolean PrefFormHandleEvent(EventPtr event);
// details form
static void DetailsFormInit(FormPtr frmDetails, Boolean firstTime);
static Boolean DetailsFormHandleEvent(EventPtr event);
static void SetDateLabel(ControlPtr ctl, DateTimePtr date);
static void SetTimeLabel(ControlPtr ctl, DateTimePtr time);
static void DTToTime(DateTimePtr dtPtr, TimePtr timePtr);
static void TimeToDT(TimePtr timePtr, DateTimePtr dtPtr);
// quick send form
static Boolean QuickSendFormHandleEvent(EventPtr event);
// receive form
static void ReceiveFormInit(FormPtr frmReceive);
static void ReceiveFormClose(FormPtr frmReceive);
static Boolean ReceiveFormHandleEvent(EventPtr event);
// send form
static void SendFormInit(FormPtr frmSend);
static void SendFormClose(FormPtr frmSend);
static Boolean SendFormHandleEvent(EventPtr event);

// events
static void EventLoop(void);

// serial
static Err RxData(void);
static Err SetSerialSettings(void);
static Err OpenPort(void);
static void ClosePort(void);
static Boolean SendData(void);

// error form
// static UInt FrmError(CharPtr s);
#define FrmError(s) (FrmCustomAlert(idAlertError, s, "", ""))
#ifdef DEBUG
// debug form
static void DebugDialog(CharPtr v1, Long v2I, Long v3I);
#endif

/*
 * global variables
 *
 */
static LdSerLibPreferenceType prefs;
static UInt serialRef;
static Boolean portClosed = true;
// form stuff
static Int currentForm = noFormLoaded;


/*
 * Opens the serial library and initializes the database
 *
 */
static Err StartApplication(void) {
        SystemPreferencesType systemPrefs;
	Err error;
		
	PrefGetPreferences(&systemPrefs);

	error = SysLibFind("Serial Library", &serialRef);
	if (error) {
	    if (error == sysErrLibNotFound) {
		error = SysLibLoad('libr', 'swrp', &serialRef);
		if (error) {
		    FrmError("gru: sysLibLoad error for serial lib");
		    return error;
		}
	    } else {
		FrmError("gru: Serial lib access error");
		return error;
	    }
	}

	get_or_create_prefs();
	FrmGotoForm(idFormLdSerLib);

	return 0;
}

/*
 * Closes the database
 *
 */

static void StopApplication (void) {
        FrmCloseAllForms();
	ClosePort();
}

/*
 * Checks for ROM version compatibility
 *
 */
static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags) {
        DWord romVersion;
	
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion) {
	        if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) == (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
		        FrmError(strErrWrongROMVersion);

			return sysErrRomIncompatible;
		}
        }

        return 0;
}

/*
 * opens the database or creates one
 *
 */
static Err get_or_create_prefs(void) {
	UInt cardNo;
	Word prefSize;
	Err error;

	// if no appInfoID exists create one
	// read preferences
	prefSize = sizeof(LdSerLibPreferenceType);
	if (PrefGetAppPreferences(idAppLdSerLib, 0, &prefs, &prefSize, true) == noPreferenceFound) {
	        prefs.version = defaultVersion;

	        prefs.bits = defaultBits;
		prefs.parity = defaultParity;
		prefs.baud = defaultBaud;
		prefs.xon = defaultXON;
		prefs.rts = defaultRTS;


		// set
		SetSerialSettings();
	}

	return 0;
}
/*
 * returns a formatted string
 *
 */
void GetSizeString(CharPtr label, ULong size, Boolean longFormat) {
        if (size < 1024) {
	        /* byte */
		StrIToA(label, size);
		if (longFormat)
			StrCat(label, strBytes);
		else
			StrCat(label, " by");
	} else {
		/* kilo byte */
	        StrIToA(label, size / 1024);
	        StrCat(label, " kb");
       }
}


/*
 * sets serial settings and saves app-prefs
 *
 */
static Err SetSerialSettings(void) {
        SerSettingsType serialSettings;
	Err error;
	
	SerGetSettings(serialRef, &serialSettings);

	serialSettings.baudRate = prefs.baud;
	serialSettings.flags = serSettingsFlagStopBits1;
	serialSettings.ctsTimeout = serDefaultCTSTimeout;

	if (prefs.bits == 7) 
		serialSettings.flags |= serSettingsFlagBitsPerChar7;
	else 
		serialSettings.flags |= serSettingsFlagBitsPerChar8;

	if (prefs.parity == 1) 
		serialSettings.flags |= serSettingsFlagParityOnM;
	else if (prefs.parity == 2) 
		serialSettings.flags |= serSettingsFlagParityOnM | serSettingsFlagParityEvenM;

	if (prefs.xon == 1) 
		serialSettings.flags |= serSettingsFlagXonXoffM;

	if (prefs.rts == 1) 
		serialSettings.flags |= serSettingsFlagCTSAutoM | serSettingsFlagRTSAutoM;
 

	if (! portClosed) {
	        error = SerSetSettings(serialRef, &serialSettings);
		if (error) {
		        SerClearErr(serialRef);
			return error;
		}
	}

	return 0;
}

/*
 * opens the serial port
 *
 */
static Err OpenPort(void) {
	Err error;

        if (! portClosed) 
		return 0;
							
	error = SerOpen(serialRef, 0, prefs.baud);
	if (error) {
		FrmError("open for serial failed gru");
		return error;
	}

	portClosed = false;							
	error = SetSerialSettings();
	if (error)
		return error;

	return 0;
}

/*
 * closes the serial port
 *
 */
static void ClosePort() {
        if (portClosed)
		return;

	SerClose(serialRef);

	portClosed = true;
}

/*
 * handles the common menu entrys
 *
 */
static Boolean FormHandleMenuEvent(EventPtr event) {
	Boolean	handled = false;

	if (event->eType == menuEvent) {
		MenuEraseStatus(0);
		handled = true;

		switch (event->data.menu.itemID) {
		case idMenuitemKeyboard:
		        SysKeyboardDialog(kbdAlpha);
		        break;
		case idMenuitemGraffiti:
			SysGraffitiReferenceDialog (referenceDefault);
			break;
		case idMenuitemPreferences:
		        FrmPopupForm(idFormPreferences);
  			break;
		case idMenuitemAbout:
		        FrmAlert(idAlertAbout);
			break;
		case idMenuitemHelp:
		        FrmHelp(idHelpLdSerLib);
			break;
		default:
		        handled = false;
		}
	}

	return handled;
}

/*
 * initializes the main form
 *
 */
static void MainFormInit(FormPtr frmMain) {
}

/*
 * Handles events for the "main" form.
 *
 */
static Boolean MainFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif

	if (event->eType == frmOpenEvent) {
	       FormPtr frmLdSerLib = FrmGetActiveForm();

	       MainFormInit(frmLdSerLib);
	       FrmDrawForm(frmLdSerLib);

	       handled = true;
	} else if (event->eType == frmUpdateEvent) {
	        MainFormInit(FrmGetActiveForm());

		handled = false;
	} else if (event->eType == ctlSelectEvent) {
	       FormPtr frmLdSerLib = FrmGetActiveForm();

	       handled = true;

   	       switch (event->data.ctlEnter.controlID) {
	       case idButtonHelp:
		       FrmHelp(idHelpLdSerLib);

		       break;
	       default:
		       handled = false;
	       }
	} else if (event->eType == menuEvent) {
		handled = FormHandleMenuEvent(event);
	}

	return handled;
}

/*
 * init preferences form 
 *
 */
static void InitPreferences(FormPtr frm) {
	ListPtr list;
	ControlPtr ctl;
			
	// speed
	list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBaud))));
	switch (prefs.baud) {
	case 1200:
		LstSetSelection(list, 0);
		break;
	case 2400:
		LstSetSelection(list, 1);
		break;
	case 4800:
		LstSetSelection(list, 2);
		break;
	case 9600:
		LstSetSelection(list, 3);
		break;
	case 19200:
		LstSetSelection(list, 4);
		break;
	case 38400:
		LstSetSelection(list, 5);
		break;
	default:
		LstSetSelection(list, 6);
	}
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idPopupBaud))));
	CtlSetLabel(ctl, LstGetSelectionText(list, LstGetSelection(list)));

	// bits
	list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBits))));
	if (prefs.bits == 7)
		LstSetSelection(list,0);
	else 
		LstSetSelection(list, 1);
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idPopupBits))));
	CtlSetLabel(ctl, LstGetSelectionText(list, LstGetSelection(list)));

	// parity
	list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListParity))));
	if (prefs.parity == 0) 
		LstSetSelection(list, 0);
	else if (prefs.parity == 1) 
		LstSetSelection(list, 1);
	else
		LstSetSelection(list, 2);
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idPopupParity))));
	CtlSetLabel(ctl, LstGetSelectionText(list, LstGetSelection(list)));
	

	// xon/xoff
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckXonXoff))));
	if (prefs.xon == 0) 
		CtlSetValue (ctl, 0);
	else 
		CtlSetValue (ctl, 1);

	// rts/cts
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckRTSCTS))));
	if (prefs.rts == 0) 
		CtlSetValue(ctl, 0);
	else 
		CtlSetValue (ctl, 1);
}

/*
 * handles the events for the preferences dialog
 *
 */			
static Boolean PrefFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif

	if (event->eType == frmOpenEvent) {
	        FormPtr formPref = FrmGetActiveForm();

		InitPreferences(formPref);
		FrmDrawForm(formPref);
		
		handled = true;
	} else if (event->eType == ctlSelectEvent) {
	        Int choice;
		ListPtr list;
		ControlPtr ctl;
		FormPtr frm;
				
		switch (event->data.ctlSelect.controlID) {
		case idButtonOkay:
			frm = FrmGetActiveForm();
							
			// save preferences
			// speed
			list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBaud))));
			choice = LstGetSelection(list);
			switch (choice) {
			case 0:
				prefs.baud = 1200;
				break;
			case 1:
				prefs.baud = 2400;
				break;
			case 2:
				prefs.baud = 4800;
				break;
			case 3:
				prefs.baud = 9600;
				break;
			case 4:
				prefs.baud = 19200;
				break;
			case 5:
				prefs.baud = 38400;
				break;
			default:
				prefs.baud = 57600;
			}

			
			// bits
			list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBits))));
			choice = LstGetSelection(list);
			if (choice == 0) 
				prefs.bits = 7;
			else 
				prefs.bits = 8;

			// parity
			list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListParity))));
			choice = LstGetSelection(list);
			if (choice == 0) 
				prefs.parity = 0;
			else if (choice == 1) 
				prefs.parity = 1;
			else 
				prefs.parity = 2;

			// xon/xoff
			ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckXonXoff))));
			if (CtlGetValue(ctl) == 0) 
				prefs.xon = 0;
			else 
				prefs.xon = 1;
					
			// rts/cts
			ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckRTSCTS))));
			if (CtlGetValue(ctl) == 0) 
				prefs.rts = 0;
			else 
				prefs.rts = 1;

			// set
			PrefSetAppPreferences(idAppLdSerLib, 0, 1, &prefs, sizeof(LdSerLibPreferenceType), true);

			FrmReturnToForm(currentForm);

			handled = true;
			break;
		case idButtonCancel:
			FrmReturnToForm(currentForm);

			handled = true;
			break;				
		}
	}

	#ifdef __CALLBACK_H__
	CALLBACK_EPILOGUE
	#endif

	return handled;
}

/*
 * main queue
 *
 */
static Boolean ApplicationHandleEvent(EventPtr event) {
        Boolean handled = false;

        if (event->eType == frmLoadEvent) {				
	        FormPtr frm;
		int formID;

	        formID = event->data.frmLoad.formID;
		frm = FrmInitForm(formID);
		FrmSetActiveForm(frm);

		switch (formID) {
		case idFormLdSerLib:
		       currentForm = formID;
		       FrmSetEventHandler(frm, MainFormHandleEvent);
		       break;
		case idFormPreferences:
		       FrmSetEventHandler(frm, PrefFormHandleEvent);
		       break;
		}
		handled = true;
	}
	return handled;
}

/*
 * handles all incoming events
 *
 */
static void EventLoop(void) {
	EventType event;
	Word error;
	
	do {
		// next event
		if (portClosed)
			EvtGetEvent(&event, evtWaitForever);
                else {
			EvtGetEvent(&event, 1); 
			EvtResetAutoOffTimer();
		}
					
		// give the system a chance to handle the event.
		if (! SysHandleEvent (&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! ApplicationHandleEvent(&event))
					FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

#ifdef DEBUG
/*
 * Displays 3 variables in an alert form
 */
static void DebugDialog(CharPtr v1, Long v2I, Long v3I) {
        Char v2[10];
        Char v3[10];

	StrIToA(v3, v3I);
	StrIToA(v2, v2I);
	FrmCustomAlert(idAlertError, v1, v2, v3);
}
#endif

/*
 * Starts the event loop
 *
 */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
	Err error;
	
	error = RomVersionCompatible(versionRequired, launchFlags);
        if (error) 
		return error;
	
	if (cmd == sysAppLaunchCmdNormalLaunch) {
		error = StartApplication();
		if (error)
			return error;

		EventLoop();
		StopApplication ();
	}

	return 0;
}

